/*
 * Copyright (c) 2006-2007, Lars Knoesel
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXOP 200
#define MAXMEM 100

	int yylex(void);
	void yyerror(const char*);

	enum {
		LOAD	= 1,
		STORE	= 2,
		ADD	= 3,
		SUB	= 4,
		READ	= 5,
		WRITE	= 6,
		GOTO	= 7,
		IF	= 8,
		DLOAD	= 9,
		ILOAD	= 10,
		ISTORE	= 11,
		MUL	= 12,
		DIV	= 13,
		END	= 14
	};
	struct {
		const char	*symbol;
		int		 value;
	} symTab[] = {
		{ "load",	LOAD },
		{ "store",	STORE },
		{ "add",	ADD },
		{ "sub",	SUB },
		{ "read",	READ },
		{ "write",	WRITE },
		{ "goto",	GOTO },
		{ "if",		IF },
		{ "dload",	DLOAD },
		{ "iload",	ILOAD },
		{ "istore",	ISTORE },
		{ "mul",	MUL },
		{ "div",	DIV },
		{ "end",	END }
	};
	struct Line {
		int	 pp;
		int	 symbol;
		int	 address;
	} linex[MAXOP];

	int pos = 0;

	int	 getSymbol(void);
	void	 saveLine(void);
	void	 simulate(void);
%}
%start lines
%token NUM NL SYM TAB UNKNOWN
%%
lines	:	/* Empty. */
	|	lines line
;

line	:	NUM NL			{ ++pos; }
|		NUM TAB SYM NL		{ ++pos; }
|		NUM TAB first NL	{ ++pos; }
|		error			{ (void) printf("Line: %d\n", pos + 1); exit(-1); } 
;

first	:       SYM TAB NUM		{ linex[pos].address = yylval; }
|		first TAB SYM TAB NUM	{ linex[pos].pp = yylval; }
;

%%
#include "lex.yy.c"

int
main(int argc, char *argv[])
{
	FILE* fp;

	(void) memset(linex, 0, sizeof(linex));

	(void) printf("MAXOP: %d\nMAXMEM: %d\n", MAXOP, MAXMEM);

	(void) yyparse();
	
	if ((fp = freopen(argv[1], "r", stdin)) == NULL) {
		(void) fprintf(stderr, "Error while opening %s, abort.", argv[1]);
		exit (-1);
	}
	simulate();

	(void) fclose(fp);

	return (0);
}
int
getSymbol(void)
{
	unsigned i;
	
	for (i = 0; i < MAXOP; ++i) {
		if (!strcmp(symTab[i].symbol, yytext)) {
			return (symTab[i].value);
		}
	}
	/* Never get here. */
	return (-1);
}
void
simulate(void)
{
	int		 mem[MAXMEM];
	int		 accu;
	int		 i;

	(void) memset(mem, 0, sizeof(mem));

	if (MAXOP < pos) {
		(void) fprintf(stderr, "MAXOP too small! Edit ram.y, abort.\n");
		exit(-1);
	}

	for (i = 0; i < MAXOP; ++i) {

		switch (linex[i].symbol)
		{
		case LOAD:	accu = mem[linex[i].address];		break;
		case READ:	(void) printf("Input: ");
				(void) scanf("%d", &accu);      	break;
		case STORE:	mem[linex[i].address] = accu;		break;
		case ADD:	accu += mem[linex[i].address];		break;
		case SUB:	accu -= mem[linex[i].address];		break;
		case GOTO:	i = linex[i].address - 1;		break;
		case DLOAD:	accu = linex[i].address;		break;
		case ILOAD:	accu = mem[mem[linex[i].address]];	break;
		case ISTORE:	mem[mem[linex[i].address]] = accu;	break;
		case MUL:	accu *= mem[linex[i].address];		break;
		case DIV:	accu /= mem[linex[i].address];		break;
		case IF:	if (accu == linex[i].address)
					i = linex[i].pp - 1;		break;
		case WRITE:	(void) printf("Output: %d\n", accu);	break;
		case END:						return;
		}
	}
	/* Never get here. */
}
